// sota staircase SideStepper
// a fast .gitignore-respecting large file finder
//
// Copyright (c) 2025 mark joshwel <mark@joshwel.co>
//
// Permission to use, copy, modify, and/or distribute this software for any
// purpose with or without fee is hereby granted.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
// SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
// IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

use std::error::Error;
use std::fs::metadata;
use std::io::Write;
use std::path::{Path, PathBuf};
use std::time::{Duration, SystemTime};
use std::{env, fs, io};

use ignore;

// const SOTA_SIDESTEP_CHUNK_SIZE: u16 = 16;
// const SOTA_SIDESTEP_MAX_WORKERS: u16 = 4;
const SOTA_SIDESTEP_LARGE_FILE_SIZE: u64 = 100000000; // 100mb

struct Behaviour {
    repo_dir_path: PathBuf,
    repo_sotaignore_path: PathBuf,
    // parallel: bool,
    // chunk_size: u16,
    // max_workers: u16,
    large_file_size: u64,
    plumbing: bool,
}

fn cli_get_behaviour() -> Result<Behaviour, Box<dyn Error>> {
    // get environment variables
    // let parallel: bool = 'get_parallel: {
    //     // future me move this to a higher block if we ever need args
    //     // anywhere else also what the hell, labeled blocks?
    //     // huh -- the community seems wishy-washy on it,
    //     // but this seems like a harmless use of em
    //     let args: Vec<String> = env::args().collect();
    //     if env::var("SOTA_SIDESTEP_PARALLEL").is_ok() {
    //         break 'get_parallel true;
    //     }
    //     if args.iter().any(|arg| arg == "--parallel") {
    //         break 'get_parallel true;
    //     }
    //     false
    // };
    // let chunk_size: u16 = match env::var("SOTA_SIDESTEP_CHUNK_SIZE") {
    //     Ok(val) => val.parse::<u16>().unwrap_or(SOTA_SIDESTEP_CHUNK_SIZE),
    //     Err(_) => SOTA_SIDESTEP_CHUNK_SIZE,
    // };
    // let max_workers: u16 = match env::var("SOTA_SIDESTEP_MAX_WORKERS") {
    //     Ok(val) => val.parse::<u16>().unwrap_or(SOTA_SIDESTEP_MAX_WORKERS),
    //     Err(_) => SOTA_SIDESTEP_MAX_WORKERS,
    // };
    let large_file_size: u64 = match env::var("SOTA_SIDESTEP_LARGE_FILE_SIZE") {
        Ok(val) => val.parse::<u64>().unwrap_or(SOTA_SIDESTEP_LARGE_FILE_SIZE),
        Err(_) => SOTA_SIDESTEP_LARGE_FILE_SIZE,
    };

    // look through args and see if the '--search-here' or '--plumbing' flags are present
    let mut search_here: bool = false;
    let mut plumbing: bool = false;
    for arg in env::args() {
        if arg == "--search-here" {
            search_here = true;
        }
        if arg == "--plumbing" {
            plumbing = true;
        }
    }

    let current_dir = env::current_dir().unwrap();

    // if we're searching here anywas, return early using the current dir
    if search_here {
        return Ok(Behaviour {
            repo_dir_path: PathBuf::from(&current_dir),
            repo_sotaignore_path: PathBuf::from(current_dir.join(".sotaignore")),
            large_file_size,
            plumbing,
        });
    }

    // else, find the repo dir
    // (go through each parent dir until one of them has a .git directory in it)
    let mut dir = current_dir.as_path();
    let mut possible_repo_dir_path: Option<&Path> = None;
    while dir.components().count() > 1 {
        // check if there's a .git directory nearby
        if dir.join(".git/").try_exists().ok() == Some(true) {
            possible_repo_dir_path = Option::from(dir);
            break;
        }

        // iterate down!
        if let Some(parent) = dir.parent() {
            dir = parent;
        } else {
            break;
        }
    }
    if possible_repo_dir_path.is_none() {
        return Err("could not find a .git repository in the current or parent directories".into());
    }
    let repo_dir_path = possible_repo_dir_path.unwrap();
    Ok(Behaviour {
        repo_dir_path: PathBuf::from(repo_dir_path),
        repo_sotaignore_path: PathBuf::from(repo_dir_path.join(".sotaignore")),
        large_file_size,
        plumbing,
    })
}

fn ss_scan_for_unignored_files(behaviour: &Behaviour) -> Vec<PathBuf> {
    // for file in ignore::WalkBuilder::new(&behaviour.repo_dir_path)
    //     .hidden(false)
    //     .build()
    //     .into_iter()
    //     .filter_map(|e| e.ok())
    // {
    //     if file
    //         .path()
    //         .starts_with(Path::new(&behaviour.repo_dir_path).join(".git/"))
    //     {
    //         continue;
    //     }
    //     if file.path().is_file() {
    //         files.push(file.into_path());
    //     }
    // }
    ignore::WalkBuilder::new(&behaviour.repo_dir_path)
        .hidden(false)
        .build()
        .filter_map(|e| e.ok())
        .filter(|file| {
            !file
                .path()
                .starts_with(Path::new(&behaviour.repo_dir_path).join(".git/"))
                && file.path().is_file()
        })
        .map(|file| file.into_path())
        .collect()
}

fn ss_check_for_large_files(behaviour: &Behaviour, files: &Vec<PathBuf>) -> Vec<PathBuf> {
    // let mut large_files: Vec<PathBuf> = Vec::new();
    // for file in files {
    //     let result = metadata(file);
    //     if let Err(_) = result {
    //         continue;
    //     }
    //     let metadata = result.unwrap();
    //     if metadata.len() >= behaviour.large_file_size {
    //         large_files.push(file.into());
    //     }
    // }
    files
        .iter()
        .filter_map(|file| {
            metadata(file)
                .ok()
                .filter(|meta| meta.len() >= behaviour.large_file_size)
                .map(|_| file.into())
        })
        .collect()
}

fn ss_write_sotaignore(behaviour: &Behaviour, large_files: &Vec<PathBuf>) -> io::Result<bool> {
    if large_files.is_empty() {
        return Ok(false);
    }

    // are we outputting to stdout for other programs?
    // do so and return true, we did write something
    if behaviour.plumbing {
        eprintln!();
        for file in large_files {
            println!("{}", file.to_str().unwrap());
        }
        return Ok(true);
    }

    let old_sotaignore = if behaviour.repo_sotaignore_path.try_exists().ok() == Some(true) {
        fs::read_to_string(&behaviour.repo_sotaignore_path)?
            .lines()
            .map(String::from)
            .collect::<Vec<String>>()
    } else {
        Vec::new()
    };

    let mut new_sotaignore = old_sotaignore.clone();
    for file in large_files {
        if let Ok(file_relative) = file.strip_prefix(&behaviour.repo_dir_path) {
            let relative_path_str = file_relative.to_string_lossy();
            if !old_sotaignore.contains(&relative_path_str.to_string()) {
                new_sotaignore.push(relative_path_str.to_string());
            }
        }
    }

    // no new changes? return, nothing has been written
    if new_sotaignore == old_sotaignore {
        return Ok(false);
    }

    // check if the sotaignore file starts with a comment
    if !new_sotaignore.is_empty() & !new_sotaignore[0].starts_with("#") {
        let header = vec![
            "# .sotaignore file generated by sota staircase ReStepper/SideStepper",
            "# anything here either can't or shouldn't be uploaded to GitHub",
            "# unless you know what you're doing, don't edit this file! >:(",
        ];
        new_sotaignore.splice(0..0, header.iter().map(|&line| line.to_string()));
    }

    let mut sotaignore_file = fs::File::create(&behaviour.repo_sotaignore_path)?;
    sotaignore_file.write_all(new_sotaignore.join("\n").as_bytes())?;
    sotaignore_file.write_all(b"\n")?;

    Ok(true)
}

fn format_elapsed_time(secs: f64) -> String {
    let hours = (secs / 3600.0).floor() as i64;
    let minutes = ((secs % 3600.0) / 60.0).floor() as i64;
    let seconds = (secs % 60.0).round() as f64;
    let secs_string: String;
    if secs > 3600.0 {
        secs_string = format!("{}h {}′ {:.1}″", hours, minutes, seconds);
    } else if secs > 60.0 {
        secs_string = format!("{}′ {:.2}″", minutes, seconds);
    } else {
        secs_string = format!("{:.3}″", secs);
    }
    secs_string
}

fn main() {
    eprintln!("\nsota staircase SideStepper v5 (i3/a5)");
    let behaviour = {
        let behaviour = cli_get_behaviour();
        // huh. pattern matching consumes the variable, so we ref (&) it. damn.
        if let Err(e) = &behaviour {
            eprintln!("critical error: {}\n", e);
            std::process::exit(1);
        }
        behaviour.unwrap()
    };
    eprintln!(
        "   repo root : {}\n .sotaignore : {}\n",
        behaviour.repo_dir_path.to_str().unwrap(),
        {
            if behaviour.plumbing {
                "(stdout)".into()
            } else {
                format!(
                    "{} ({})",
                    behaviour.repo_sotaignore_path.to_str().unwrap(),
                    {
                        if behaviour.repo_sotaignore_path.try_exists().ok() == Some(true) {
                            "exists"
                        } else {
                            "non-existent"
                        }
                    }
                )
            }
        },
    );

    let zero_duration = Duration::new(0, 0);
    let all = SystemTime::now();

    eprint!("1/3   scanning repository... ");
    let now = SystemTime::now();
    let files = ss_scan_for_unignored_files(&behaviour);
    eprintln!(
        "done in {} (found {})",
        format_elapsed_time(now.elapsed().unwrap_or(zero_duration).as_secs_f64()),
        files.len()
    );

    eprint!("2/3   finding large files... ");
    let now = SystemTime::now();
    let large_files = ss_check_for_large_files(&behaviour, &files);
    eprintln!(
        "done in {} (found {})",
        format_elapsed_time(now.elapsed().unwrap_or(zero_duration).as_secs_f64()),
        large_files.len()
    );

    eprint!("3/3   writing .sotaignore file... ");
    match ss_write_sotaignore(&behaviour, &large_files) {
        Ok(true) => {
            eprintln!(
                "{}",
                if behaviour.plumbing {
                    "done (to stdout)"
                } else {
                    "done"
                }
            );
        }
        Ok(false) => {
            eprintln!("skipped")
        }
        Err(e) => {
            eprintln!("error ({})", e)
        }
    }

    eprintln!(
        "\n--- done! took {} ″~ ☆*: .｡. o(≧▽≦)o .｡.:*☆ ---\n",
        format_elapsed_time(all.elapsed().unwrap_or(zero_duration).as_secs_f64())
    );
}
